# -*- coding: utf-8 -*-
#
#  Tumgreyspf
#  Copyright © 2004-2005, Sean Reifschneider, tummy.com, ltd.
#
#  pypolicyd-spf changes
#  Copyright © 2007-16 Scott Kitterman <scott@kitterman.com>
'''
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
'''

import syslog
import os
import sys
import re
import stat
import socket


#  default values
defaultConfigData = {
        'debugLevel' : 1,
        'HELO_reject' : 'Fail',
        'Mail_From_reject' : 'Fail',
        'PermError_reject' : 'False',
        'TempError_Defer'  : 'False',
        'skip_addresses' : '127.0.0.0/8,::ffff:127.0.0.0/104,::1',
        'TestOnly' : 1,
        'SPF_Enhanced_Status_Codes' : 'Yes',
        'Header_Type' : 'SPF',
        'Hide_Receiver' : 'Yes',
        'Authserv_Id' : 'HOSTNAME',
        'Lookup_Time' : 20,
        'Whitelist_Lookup_Time' : 10,
        'Void_Limit' : 2,
        'Reason_Message' : 'Message {rejectdefer} due to: {spf}. Please see {url}',
        'No_Mail' : False,
        'Mock' : False,
	# For milter front end
        'Socket': 'local:/run/pyspf-milter/pyspf-milter.sock',
        'PidFile': '/run/pyspf-milter/pyspf-milter.pid',
        'UserID': 'pyspf-milter',
        'UMask': 7,
        'InternalHosts': '127.0.0.1',
        'IntHosts': False,
        'MacroListVerify': '',
        }


#################################
class ConfigException(Exception):
    '''Exception raised when there's a configuration file error.'''
    pass


####################################################################
def _processConfigFile(filename = None, config = None, useSyslog = 1,
        useStderr = 0):
    '''Load the specified config file, exit and log errors if it fails,
    otherwise return a config dictionary.'''

    import spf_engine.policydspfsupp
    if config == None: config = spf_engine.policydspfsupp.defaultConfigData
    if filename != None:
        try:
            _readConfigFile(filename, config)
        except Exception as e:
            if useSyslog:
                syslog.syslog(e.args[0])
            if useStderr:
                sys.stderr.write('%s\n' % e.args[0])
            sys.exit(1)
    return(config)


#################
class ExceptHook:
    def __init__(self, useSyslog = 1, useStderr = 0):
        self.useSyslog = useSyslog
        self.useStderr = useStderr

    def __call__(self, etype, evalue, etb):
        import traceback
        import sys
        tb = traceback.format_exception(*(etype, evalue, etb))
        for line in tb:
            if self.useSyslog:
                syslog.syslog(line)
            if self.useStderr:
                sys.stderr.write(line)

####################
def _setExceptHook():
    sys.excepthook = ExceptHook(useSyslog = 1, useStderr = 1)


def _make_authserv_id(as_id):
    """Determine Authserv_Id if needed"""
    if as_id == 'HOSTNAME':
        as_id = socket.gethostname()
    return as_id

###############################################################
commentRx = re.compile(r'^(.*)#.*$')
def _readConfigFile(path, configData = None, configGlobal = {}):
    '''Reads a configuration file from the specified path, merging it
    with the configuration data specified in configData.  Returns a
    dictionary of name/value pairs based on configData and the values
    read from path.'''

    debugLevel = configGlobal.get('debugLevel', 0)
    if debugLevel >= 5: syslog.syslog('readConfigFile: Loading "%s"' % path)
    if configData == None: configData = {}
    nameConversion = {
            'debugLevel' : int,
            'HELO_reject' : str,
            'Mail_From_reject' : str,
            'PermError_reject' : str,
            'TempError_Defer' : str,
            'Mail_From_pass_restriction' : str,
            'HELO_pass_restriction' : str,
            'Prospective' : str,
            'Whitelist' : str,
            'skip_addresses': str,
            'HELO_Whitelist': str,
            'Domain_Whitelist' : str,
            'Domain_Whitelist_PTR': str,
            'No_Mail': str,
            'Reject_Not_Pass_Domains' : str,
            'Per_User' : str,
            'TestOnly' : int,
            'defaultSeedOnly' : int,
            'SPF_Enhanced_Status_Codes' : str,
            'Header_Type' : str,
            'Hide_Receiver' : str,
            'Authserv_Id' : str,
            'Lookup_Time' : int,
            'Whitelist_Lookup_Time': int,
            'Void_Limit'  : int,
            'Reason_Message' : str,
            'Mock' : bool,
            # For milter front end
            'Socket': str,
            'PidFile': str,
            'UserID': str,
            'UMask': 'int',
            'InternalHosts': str,
            'IntHosts': 'bool',
            'MacroListVerify': str,
            }


    #  check to see if it's a file
    try:
        mode = os.stat(path)[0]
    except OSError as e:
        if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR,'ERROR stating "%s": %s' % ( path, e.strerror ))
        return(configData)
    if not stat.S_ISREG(mode):
        if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR,'ERROR: is not a file: "%s", mode=%s' % ( path, oct(mode) ))
        return(configData)

    #  load file
    fp = open(path, 'r')
    while 1:
        line = fp.readline()
        if not line: break

        #  parse line
        line = (line.split('#', 1)[0]).strip()
        if debugLevel >= 3: syslog.syslog(str(line))
        if not line: continue
        data = [q.strip() for q in line.split('=')]
        if len(data) != 2:
            if len(data) == 1:
                if debugLevel >= 1:
                    syslog.syslog('Configuration item "%s" not defined in file "%s"'
                        % ( line, path ))
            else:
                syslog.syslog('ERROR parsing line "%s" from file "%s"'
                    % ( line, path ))
            continue
        name, value = data

        #  check validity of name
        conversion = nameConversion.get(name)
        if conversion == None:
            if debugLevel >= 0: syslog.syslog('ERROR: Unknown name "%s" in file "%s"' % ( name, path ))
            continue

        if debugLevel >= 5: syslog.syslog('readConfigFile: Found entry "%s=%s"'
                % ( name, value ))
        configData[name] = conversion(value)
    fp.close()
    try:
        if debugLevel >= 5: syslog.syslog('Authserv_Id before: {0}'.format(configData['Authserv_Id']))
        configData['Authserv_Id'] = _make_authserv_id(configData['Authserv_Id'])
        if debugLevel >= 5: syslog.syslog('Authserv_Id after: {0}'.format(configData['Authserv_Id']))
    except:
        pass
    
    return(configData)




#! /usr/bin/python3
# Derived from dkim-milter.py code:
# Author: Stuart D. Gathman <stuart@bmsi.com>
# Copyright 2007 Business Management Systems, Inc.
# This code is under GPL.  See COPYING for details.
# and:
# dkimpy-milter: A DKIM signing/verification Milter application
# Author: Scott Kitterman <scott@kitterman.com>
# Copyright 2018 Scott Kitterman
# spf-milter.py:
# Author: Scott Kitterman <scott@kitterman.com>
# Copyright 2019 Scott Kitterman
"""    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import sys
import syslog
import Milter
import authres
import os
import tempfile
import re
from Milter.utils import parse_addr, parseaddr
import spf_engine
import spf_engine.policydspfsupp as config
from spf_engine.util import drop_privileges
from spf_engine.policydspfsupp import _setExceptHook
from spf_engine.util import write_pid
from spf_engine.util import fold

__version__ = "2.9.3"
FWS = re.compile(r'\r?\n[ \t]+')


class spfMilter(Milter.Base):
    "Milter to check SPF.  Each connection gets its own instance."

    def __init__(self):
        self.mailfrom = None
        self.id = Milter.uniqueID()
        self.instance_dict = {'0':'init',}
        self.instance_dict.clear()
        self.data = {}

    @Milter.noreply
    def connect(self, hostname, unused, hostaddr):
        self.internal_connection = False
        self.external_connection = False
        self.hello_name = None
        # sometimes people put extra space in sendmail config, so we strip
        self.receiver = self.getsymval('j').strip()
        try:
            self.Authserv_Id = milterconfig['Authserv_Id']
        except:
            self.Authserv_Id = self.receiver
        if hostaddr and len(hostaddr) > 0:
            ipaddr = hostaddr[0]
            if milterconfig['IntHosts']:
                if milterconfig['IntHosts'].match(ipaddr):
                    self.internal_connection = True
        else:
            ipaddr = ''
        self.connectip = ipaddr
        self.data['client_address'] = ipaddr
        self.data['helo_name'] = hostname
        self.data['recipient'] = '<UNKNOWN>'
        if milterconfig.get('MacroList') and not self.internal_connection:
            macrolist = milterconfig.get('MacroList')
            for macro in macrolist:
                macroname = macro.split('|')[0]
                macroname = '{' + macroname + '}'
                macroresult = self.getsymval(macroname)
                if ((len(macro.split('|')) == 1 and macroresult) or macroresult
                        in macro.split('|')[1:]):
                    self.external_connection = True
        if self.internal_connection:
            connecttype = 'INTERNAL'
        else:
            connecttype = 'EXTERNAL'
        if milterconfig.get('debugLevel') >= 1:
            syslog.syslog("connect from {0} at {1} {2}"
                          .format(hostname, hostaddr, connecttype))
        return Milter.CONTINUE

    # multiple messages can be received on a single connection
    # envfrom (MAIL FROM in the SMTP protocol) seems to mark the start
    # of each message.
    def envfrom(self, f, *str):
        if milterconfig.get('debugLevel') >= 2:
            syslog.syslog("mail from: {0} {1}".format(f, str))
        self.mailfrom = f
        t = parse_addr(f)
        if len(t) == 2:
            t[1] = t[1].lower()
            domain = t[1]
        else:
            domain = 'localhost.localdomain'
        self.canon_from = '@'.join(t)
        self.arheaders = []
        self.arresults = []
        self.data['sender'] = self.canon_from 
        if (not self.internal_connection or self.external_connection) and self.connectip:
            return self.check_spf()
        return Milter.CONTINUE

    @Milter.noreply
    def header(self, name, val):
        lname = name.lower()
        if lname == 'authentication-results':
            self.arheaders.append(val)
        return Milter.CONTINUE

    @Milter.noreply
    def eoh(self):
        self.bodysize = 0
        return Milter.CONTINUE

    @Milter.noreply
    def body(self, chunk):        # copy body to temp file
        return Milter.CONTINUE

    def eom(self):
        if self.arresults:
            h = results=self.arresults[0]
            h = fold(str(h))
            if milterconfig.get('debugLevel') >= 2:
                syslog.syslog(str(h))
            name, val = str(h).split(': ', 1)
            self.addheader(name, val, 0)
        return Milter.CONTINUE


    def check_spf(self):
        peruser = False
        perusermilterconfig = []
        rejectAction = ''
        deferAction = ''
        if not self.data.get('recipient'):
            self.data['recipient'] = 'none'
        if milterconfig.get('debugLevel') >= 3: syslog.syslog('Config: %s' % str(milterconfig))
        #  run the checkers  {{{3
        checkerValue = None
        checkerReason = None
        checkerValue, checkerReason, self.instance_dict, iserror = \
                spf_engine._spfcheck(self.data, self.instance_dict, milterconfig,
                peruser, perusermilterconfig)

        TestOnly = milterconfig.get('TestOnly')
        if TestOnly == 0 and checkerValue != 'prepend':
            checkerValue = None
            checkerReason = None

        if milterconfig.get('SPF_Enhanced_Status_Codes') == 'No':
            rejectAction = ''
            deferAction = 'defer_if_permit'
        else:
            if iserror:
                rejectAction = '5.7.24'
                deferAction = '4.7.24'
            else:
                rejectAction = '5.7.23'
                deferAction = '4.7.24'

        #  handle results  {{{3
        if milterconfig.get('Header_Type') != 'None' and checkerValue != 'reject':
            if milterconfig.get('debugLevel') >= 3: syslog.syslog('{0} {1}'.format('Header text', checkerReason))
            self.arresults.append(checkerReason)
        if milterconfig.get('debugLevel') >= 3: syslog.syslog('Action: {0}: Text: {1} Reject action: {2}'.format(checkerValue, checkerReason, rejectAction))

        if checkerValue == 'reject':
            if milterconfig.get('debugLevel') >= 1: syslog.syslog('{0} {1}'.format(rejectAction, checkerReason))
            self.setreply(str(550), rejectAction, checkerReason)
            return Milter.REJECT

        elif checkerValue == 'prepend':
            if milterconfig.get('Prospective'):
                return Milter.CONTINUE
            else:
                if milterconfig.get('Header_Type') != 'None':
                    if milterconfig.get('debugLevel') >= 1: syslog.syslog('prepend {0}'.format(checkerReason))
                else:
                    if milterconfig.get('debugLevel') >= 1: syslog.syslog('Header field not prepended: {0}'.format(checkerReason))
                return Milter.CONTINUE

        elif checkerValue == 'defer':
            if milterconfig.get('debugLevel') >= 1: syslog.syslog('{0} {1}'.format(deferAction, checkerReason))
            self.setreply(str(450), deferAction, checkerReason)
            return Milter.TEMPFAIL

        elif checkerValue == 'result_only':
            return Milter.SKIP
        else:
            return Milter.CONTINUE


def main():
    # Ugh, but there's no easy way around this.
    global milterconfig
    configFile = '/usr/local/etc/python-policyd-spf/policyd-spf.conf'
    if len(sys.argv) > 1:
        if sys.argv[1] in ('-?', '--help', '-h'):
            print('usage: pyspf-milter [<configfilename>]')
            sys.exit(1)
        configFile = sys.argv[1]
    milterconfig = config._processConfigFile(filename=configFile)
    # Unlike policyd, milter interface only suppports authres
    if milterconfig.get('Header_Type') != 'None':
        milterconfig['Header_Type'] = 'AR'
    # Per user configurations not supported on milter interface
    milterconfig['Per_User'] = False
    facility = eval("syslog.LOG_{0}".format('MAIL'))
    syslog.openlog(os.path.basename(sys.argv[0]), syslog.LOG_PID, facility)
    _setExceptHook()
    pid = write_pid(milterconfig)
    Milter.factory = spfMilter
    Milter.set_flags(Milter.CHGHDRS + Milter.ADDHDRS)
    miltername = 'pyspf-filter'
    socketname = milterconfig.get('Socket')
    syslog.syslog('pyspf-milter started:{0} user:{1}'
                  .format(pid, milterconfig.get('UserID')))
    sys.stdout.flush()
    drop_privileges(milterconfig)
    Milter.runmilter(miltername, socketname, 240)

if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#  Check SPF results and provide recommended action back to Postfix.
#
#  Tumgreyspf source
#  Copyright © 2004-2005, Sean Reifschneider, tummy.com, ltd.
#  <jafo@tummy.com>
#
#  pypolicyd-spf
#  Copyright © 2007-16, Scott Kitterman <scott@kitterman.com>
'''
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
'''

def main():
    __version__ = "2.9.3"

    import syslog
    import os
    import socket
    import sys
    import re

    import spf_engine
    import spf_engine.policydspfsupp as policydspfsupp
    # import policydspfuser - Only imported if per-user settings are activated for
    # efficiency.

    if int(sys.version_info.major) < 3 or (int(sys.version_info.major) == 3 and \
            int(sys.version_info.minor) < 3):
        raise ImportError("Python 3.3 or later is required")

    syslog.openlog(os.path.basename(sys.argv[0]), syslog.LOG_PID, syslog.LOG_MAIL)
    policydspfsupp._setExceptHook()

    #  load config file  {{{1
    #  Default location:
    configFile = '/etc/python-policyd-spf/policyd-spf.conf'
    if len(sys.argv) > 1:
        if sys.argv[1] in ( '-?', '--help', '-h' ):
            print('usage: policyd-spf [<configfilename>]')
            sys.exit(1)
        configFile = sys.argv[1]

    configGlobal = policydspfsupp._processConfigFile(filename = configFile)

    #  loop reading data  {{{1
    debugLevel = configGlobal.get('debugLevel', 1)
    if debugLevel >= 3: syslog.syslog('Starting')
    instance_dict = {'0':'init',}
    instance_dict.clear()
    data = {}
    lineRx = re.compile(r'^\s*([^=\s]+)\s*=(.*)$')
    while 1:
        # Python readline assumes ascii here, but sometimes it's not
        lineraw = sys.stdin.buffer.readline()
        line = lineraw.decode('UTF-8',errors='replace')
        if not line: break
        line = line.rstrip()
        if debugLevel >= 4: syslog.syslog('Read line: "%s"' % line)

        #  end of entry  {{{2
        if not line:
            peruser = False
            peruserconfigData = []
            rejectAction = ''
            deferAction = ''
            if debugLevel >= 4: syslog.syslog('Found the end of entry')
            configData = dict(list(configGlobal.items()))
            if configData.get('defaultSeedOnly') is not None:
                syslog.syslog(syslog.LOG_ERR,'WARNING: Deprecated Config Option defaultSeedOnly in use in: {0}'.format(configFile ))

            if configData.get('Mock'):
                sys.stdout.write('action=dunno mock header field that should be ignored\n\n')
                continue
            if not data.get('recipient'):
                data['recipient'] = 'none'
            if configData.get('Per_User'):
                import spf_engine.policydspfuser
                peruserconfigData, peruser = spf_engine.policydspfuser._datacheck(configData, data.get('recipient'))
                if debugLevel >= 2 and peruser: syslog.syslog('Per user configuration data in use for: %s' \
                    % str(data.get('recipient')))
            if configData.get('Hide_Receiver') != 'No':
                data['recipient'] = '<UNKNOWN>'
            if debugLevel >= 3: syslog.syslog('Config: %s' % str(configData))
            #  run the checkers  {{{3
            checkerValue = None
            checkerReason = None
            checkerValue, checkerReason, instance_dict, iserror = \
                    spf_engine._spfcheck(data, instance_dict, configData,
                    peruser, peruserconfigData)

            if not peruser:
                TestOnly = configData.get('defaultSeedOnly')
                TestOnly = configData.get('TestOnly')
            else:
                TestOnly = peruserconfigData.get('defaultSeedOnly')
                TestOnly = peruserconfigData.get('TestOnly')
            if TestOnly == 0 and checkerValue != 'prepend':
                checkerValue = None
                checkerReason = None

            if configData.get('SPF_Enhanced_Status_Codes') == 'No':
                rejectAction = '550'
                deferAction = 'defer_if_permit'
            else:
                if iserror:
                    rejectAction = '550 5.7.24'
                    deferAction = 'defer_if_permit 4.7.24'
                else:
                    rejectAction = '550 5.7.23'
                    deferAction = 'defer_if_permit 4.7.24'

            #  handle results  {{{3
            if debugLevel >= 3: syslog.syslog('Action: {0}: Text: {1} Reject action: {2}'.format(checkerValue, checkerReason, rejectAction))
            if checkerValue == 'reject':
                if debugLevel >= 1: syslog.syslog('{0} {1}'.format(rejectAction, checkerReason))
                sys.stdout.write('action={0} {1}\n\n'.format(rejectAction, checkerReason))

            elif checkerValue == 'prepend':
                if configData.get('Prospective'):
                    sys.stdout.write('action=dunno\n\n')
                else:
                    if configData.get('Header_Type') != 'None':
                        if debugLevel >= 1: syslog.syslog('prepend {0}'.format(checkerReason))
                        try:
                            sys.stdout.write('action=prepend %s\n\n' % checkerReason)
                        except UnicodeEncodeError:
                            sys.stdout.write('action=prepend %s\n\n' % str(checkerReason.encode("UTF-8"))[1:].strip("'"))
                    else:
                        if debugLevel >= 1: syslog.syslog('Header field not prepended: {0}'.format(checkerReason))
                        sys.stdout.write('action=dunno\n\n')

            elif checkerValue == 'defer':
                if debugLevel >= 1: syslog.syslog('{0} {1}'.format(deferAction, checkerReason))
                try:
                    sys.stdout.write('action={0} {1}\n\n'.format(deferAction, checkerReason))
                except UnicodeEncodeError:
                    sys.stdout.write('action={0} {1}\n\n'.format(deferAction, str(checkerReason.encode("UTF-8"))[1:].strip("'")))

            elif checkerValue == 'warn':
                try:
                   sys.stdout.write('action=warn %s\n\n' % checkerReason)
                except UnicodeEncodeError:
                    sys.stdout.write('action=warn %s\n\n' % str(checkerReason.encode("UTF-8"))[1:].strip("'"))

            elif checkerValue == 'result_only':
                if debugLevel >= 1: syslog.syslog('result_only {0}'.format(checkerReason))
                try:
                    sys.stdout.write('action=%s\n\n' % checkerReason)
                except UnicodeEncodeError:
                    sys.stdout.write('action=%s\n\n' % str(checkerReason.encode("UTF-8"))[1:].strip("'"))

            else:
                sys.stdout.write('action=dunno\n\n')

            #  end of record  {{{3
            sys.stdout.flush()
            data = {}
            continue

        #  parse line  {{{2
        m = lineRx.match(line)
        if not m: 
            if debugLevel >= 0: syslog.syslog('ERROR: Could not match line "%s"' % line)
            continue

        #  save the string  {{{2
        key = m.group(1)
        value = m.group(2)
        if key not in [ 'protocol_state', 'protocol_name', 'queue_id' ]:
            value = value.lower()
        data[key] = value

    if debugLevel >= 3: syslog.syslog('Normal exit')



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#  Check SPF results and provide recommended action back to Postfix.
#
#  Tumgreyspf source
#  Copyright © 2004-2005, Sean Reifschneider, tummy.com, ltd.
#  <jafo@tummy.com>
#
#  pypolicyd-spf
#  Copyright © 2007-16, Scott Kitterman <scott@kitterman.com>
'''
   This code is dual licensed:

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   and

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''

__version__ = "2.9.0"

import syslog
import os
import sys
import re
import socket
import ipaddress # needs python3.3, python2 backport doesn't work
# import random - Only needed for testing, so imported in line if needed.

import spf
import authres

import spf_engine.policydspfsupp
# import policydspfuser - Only imported if per-user settings are activated for
# efficiency.

if int(sys.version_info.major) < 3 or (int(sys.version_info.major) == 3 and \
    int(sys.version_info.minor) < 3):
    raise ImportError("Python 3.3 or later is required")
maj, minor, micro = spf.__version__.split(':')[0].split('.')
if int(micro) < 9:
    raise ImportError("At least pyspf 2.0.9 is required.")

syslog.openlog(os.path.basename(sys.argv[0]), syslog.LOG_PID, syslog.LOG_MAIL)
spf_engine.policydspfsupp._setExceptHook()

#############################################
def _cidrmatch(ip, netwrk):
    """Match connect IP against a CIDR network of other IP addresses."""

    address = ipaddress.ip_address(ip)
    try:
        network = ipaddress.IPv4Network(netwrk)
    except ipaddress.AddressValueError:
        network = ipaddress.IPv6Network(netwrk)
    return network.__contains__(address)

def _ipmatch(ip1, ip2):
    """Non-string matching of two IP addresses"""
    addr1 = ipaddress.ip_address(ip1)
    addr2 = ipaddress.ip_address(ip2)
    if addr1 == addr2:
        return True
    else:
        return False

def _get_rdns_lookup(ip):
    address = ipaddress.ip_address(ip)
    if isinstance(address, ipaddress.IPv4Address):
        components = address.exploded.split('.')
        return '.'.join(components[::-1]) + '.in-addr.arpa'    
    else:
        components = address.exploded.split(':')
        return '.'.join('.'.join(list(quad)) for quad in components)[::-1] + '.ip6.arpa'
#############################################
def _get_resultcodes(data, configData, scope):
    # Parse config options for SPF results to correct Posftix actions
    debugLevel = configData.get('debugLevel', 1)
    actions = {'defer':[], 'reject':[], 'prepend':[]}
    local = {'local_helo': False, 'local_mfrom': False}
    unused_results = ['Pass', 'None', 'Neutral', 'Softfail', 'Fail', 'Temperror', 'Permerror'] 
    helo_policy = ''
    mfrom_policy = ''
    reject_domain_list = []
    sender = data.get('sender')
    helo = data.get('helo_name')
    if debugLevel >= 5: syslog.syslog('_get_resultcodes: scope: {0}, Reject_Not_Pass_Domains: {1}, helo_policy: {2}, mfrom_policy: {3}'.format(scope, configData.get('Reject_Not_Pass_Domains'), configData.get('HELO_reject'),configData.get('Mail_From_reject')))
    if configData.get('Reject_Not_Pass_Domains'):
        reject_domains = (str(configData.get('Reject_Not_Pass_Domains')))
        reject_domain_list = reject_domains.split(',')
        if "@" in sender:
            sender_domain = sender.split('@', 1)[1]
        else:
            sender_domain = ''

    if scope == 'helo':
        helo_policy = configData.get('HELO_reject')
        if spf.domainmatch(reject_domain_list, helo):
            helo_policy = 'SPF_Not_Pass'
            local['local_helo'] = True
        if helo_policy == 'SPF_Not_Pass':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
                unused_results.remove('Softfail')
                actions['reject'].append('Softfail')
                unused_results.remove('Neutral')
                actions['reject'].append('Neutral')
            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: HELO_reject')
        elif helo_policy == 'Softfail':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
                unused_results.remove('Softfail')
                actions['reject'].append('Softfail')

            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: HELO_reject')
        elif helo_policy == 'Fail' or helo_policy == 'Null':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: HELO_reject')
        if debugLevel >= 5: syslog.syslog('Scope helo unused results: {0}'.format(unused_results))
    if scope == 'mfrom':
        mfrom_policy = configData.get('Mail_From_reject')
        if "@" in sender:
            sender_domain = sender.split('@', 1)[1]
        else:
            sender_domain = ''
        if spf.domainmatch(reject_domain_list, sender_domain):
            mfrom_policy = 'SPF_Not_Pass'
            local['local_mfrom'] = True
        if mfrom_policy == 'SPF_Not_Pass':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
                unused_results.remove('Softfail')
                actions['reject'].append('Softfail')
                unused_results.remove('Neutral')
                actions['reject'].append('Neutral')
            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: Mail_From_reject')
        elif mfrom_policy == 'Softfail':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
                unused_results.remove('Softfail')
                actions['reject'].append('Softfail')

            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: Mail_From_reject')
        elif mfrom_policy == 'Fail':
            try:
                unused_results.remove('Fail')
                actions['reject'].append('Fail')
            except:
                if debugLevel >= 2: syslog.syslog('Configuration File parsing error: Mail_From_reject')
        if debugLevel >= 5: syslog.syslog('Scope mfrom unused results: {0}'.format(unused_results))

    if (helo_policy == 'False' and scope == 'helo') or (mfrom_policy == 'False' and scope == 'mfrom'):
        for result in unused_results:
            actions['prepend'].append(result)
        if debugLevel >= 5: syslog.syslog('{0} policy false results: actions: {1} local {2}'.format(scope,actions,local))
        return(actions, local)
    if configData.get('TempError_Defer') == 'True':
        actions['defer'].append('Temperror')
        unused_results.remove('Temperror')
    if configData.get('PermError_reject') == 'True':
        actions['reject'].append('Permerror')
        unused_results.remove('Permerror')
    for result in unused_results:
        actions['prepend'].append(result)
    if debugLevel >= 5: syslog.syslog('{0} policy true results: actions: {1} local {2}'.format(scope,actions,local))
    return(actions, local)
#############################################
def _spfbypass(data, check_type, configData):
    debugLevel = configData.get('debugLevel', 1)
    if configData.get(check_type):
        if configData.get('Prospective'):
            ip = configData.get('Prospective')
        else:
            ip = data.get('client_address')
        bypass_list = (str(configData.get(check_type)))
        bypass_list_list = bypass_list.split(',')
        if check_type == 'skip_addresses' or check_type == 'Whitelist':
            if debugLevel >= 4: syslog.syslog ("{0} enabled.".format(check_type))
            for skip_network in bypass_list_list:
                try:
                    if _cidrmatch(ip, skip_network):
                        if check_type == 'skip_addresses':
                            comment = 'SPF check N/A for local connections - '
                        else:
                            comment = 'SPF skipped for whitelisted relay - '
                        if 'AR' in configData.get('Header_Type'):
                            try:
                                Header = str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                                    results = [authres.NoneAuthenticationResult(comment = \
                                    '{0} client-ip={1}; helo={2}; envelope-from={3}; receiver={4}'.format(comment, data.get('client_address', '<UNKNOWN>'),\
                                        data.get('helo_name', '<UNKNOWN>'), data.get('sender', '<UNKNOWN>'), data.get('recipient', '<UNKNOWN>'))
                                        )]))
                            except TypeError as x:
                                if str(x).startswith('sequence item 2'):
                                    if debugLevel >= 0:  syslog.syslog(syslog.LOG_ERR, 'Authserv_Id not set for authentication results header - invalid configuration.')
                                    raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                        else:
                            Header = ('X-Comment: ' + comment + 'client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                                % ( data.get('client_address', '<UNKNOWN>'),
                                    data.get('helo_name', '<UNKNOWN>'),
                                    data.get('sender', '<UNKNOWN>'),
                                    data.get('recipient', '<UNKNOWN>'),
                                    ))
                        if debugLevel >= 3: syslog.syslog(Header)
                        return (True, Header)
                except ValueError as msg:
                    Message = 'ERROR: {0} in {1} not IP network.  Message: {2}. Aborting whitelist processing.'.format(skip_network, check_type, msg)
                    if debugLevel >= 0: syslog.syslog(Message)
                    return (False, 'None')
            return (False, 'None')
        if check_type == 'Domain_Whitelist':
            if debugLevel >= 4: syslog.syslog ("Domain Whitelist enabled.")
            for domain in bypass_list_list:
                res = spf.check2(ip, domain, domain, querytime=configData.get('Whitelist_Lookup_Time'))
                domain_res = [res[0], res[1]]
                domain_res[0] = domain_res[0].lower()
                domain_res[0] = domain_res[0].capitalize()
                if domain_res[0] == 'Pass':
                    comment = 'SPF skipped for whitelisted relay domain - '
                    if 'AR' in configData.get('Header_Type'):
                        try:
                            Header = str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                                results = [authres.NoneAuthenticationResult(comment = \
                                '{0} client-ip={1}; helo={2}; envelope-from={3}; receiver={4}'.format(comment, data.get('client_address', '<UNKNOWN>'),\
                                    data.get('helo_name', '<UNKNOWN>'), data.get('sender', '<UNKNOWN>'), data.get('recipient', '<UNKNOWN>'))
                                    )]))
                        except TypeError as x:
                            if str(x).startswith('sequence item 2'):
                                if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR, 'Authserv_Id not set for authentication results header - invalid configuration.')
                                raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                    else:
                        Header = ('X-Comment: ' + comment + 'client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                            % ( data.get('client_address', '<UNKNOWN>'),
                                data.get('helo_name', '<UNKNOWN>'),
                                data.get('sender', '<UNKNOWN>'),
                                data.get('recipient', '<UNKNOWN>'),
                                ))
                    if debugLevel >= 3: syslog.syslog(Header)
                    return (True, Header)
            return (False, 'None')
        if check_type == 'Domain_Whitelist_PTR':
            if debugLevel >= 4: syslog.syslog ("PTR Domain Whitelist enabled.")
            try:
                # Try a reverse DNS lookup first and try and match against the domain list that way
                rDNSResults = spf.DNSLookup (_get_rdns_lookup(ip), 'ptr', timeout=configData.get('Whitelist_Lookup_Time'))
                if (len (rDNSResults) > 0):
                    rDNSName = rDNSResults [0][1]
                else:
                    # Reverse lookup didn't find any records, so don't continue with the check
                    rDNSName = None
            except spf.TempError as e:
                # DNS timeout - continue with the base SPF check.
                rDNSName = None
            if (rDNSName is not None):
                for domain in bypass_list_list:
                    if (rDNSName.endswith (domain)):
                        comment = 'SPF skipped for PTR whitelisted relay domain - '
                        if 'AR' in configData.get('Header_Type'):
                            try:
                                Header = str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                                    results = [authres.NoneAuthenticationResult(comment = \
                                    '{0} client-ip={1}; helo={2}; envelope-from={3}; receiver={4}'.format(comment, data.get('client_address', '<UNKNOWN>'),\
                                        data.get('helo_name', '<UNKNOWN>'), data.get('sender', '<UNKNOWN>'), data.get('recipient', '<UNKNOWN>'))
                                        )]))
                            except TypeError as x:
                                if str(x).startswith('sequence item 2'):
                                    if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR, 'Authserv_Id not set for authentication results header - invalid configuration.')
                                    raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                        else:
                            Header = ('X-Comment: ' + comment + 'client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                                % ( data.get('client_address', '<UNKNOWN>'),
                                    data.get('helo_name', '<UNKNOWN>'),
                                    data.get('sender', '<UNKNOWN>'),
                                    data.get('recipient', '<UNKNOWN>'),
                                    ))
                        if debugLevel >= 3: syslog.syslog(Header)
                        return (True, Header)
            return (False, 'None')
        if check_type == 'HELO_Whitelist':
            if debugLevel >= 4: syslog.syslog ("HELO Domain Whitelist enabled.")
            # Collect the IP addresses from the A/AAAA records of all HELO_Whitelist domains
            connectip = ipaddress.ip_address(ip)
            answerlist = []
            for domain in bypass_list_list:
                if domain == data.get('helo_name'):
                    try:
                        if debugLevel >= 5: syslog.syslog ("HELO Domain Whitelist domain match: {0}.".format(domain))
                        if isinstance(connectip,ipaddress.IPv4Address):
                            if debugLevel >= 5: syslog.syslog ("HELO Domain Whitelist domain: {0} is IPv4: {1}.".format(domain,connectip))
                            # DNS A lookup for IPv4 first: If A doesn't match connect IP, then skip
                            DNSAResults = spf.DNSLookup (domain, 'a', timeout=configData.get('Whitelist_Lookup_Time'))
                            if (len (DNSAResults) > 0):
                                for answer in DNSAResults:
                                    answerlist.append(answer[1])
                        if isinstance(connectip,ipaddress.IPv6Address):
                            if debugLevel >= 5: syslog.syslog ("HELO Domain Whitelist domain: {0} is IPv6: {1}.".format(domain,connectip))
                            # DNS AAAA lookup for IPv6
                            DNSAAAAResults = spf.DNSLookup (domain, 'aaaa', timeout=configData.get('Whitelist_Lookup_Time'))
                            if (len (DNSAAAAResults) > 0):
                                for answer in DNSAAAAResults:
                                    answerlist.append(answer[1])
                    except spf.TempError as e:
                        # DNS timeout - continue with the base SPF check.
                        pass
            if (answerlist is not None):
                for ipans in answerlist:
                    if _ipmatch(ipans,ip):
                        comment = 'SPF skipped for whitelisted HELO domain - '
                        if 'AR' in configData.get('Header_Type'):
                            try:
                                Header = str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                                    results = [authres.NoneAuthenticationResult(comment = \
                                    '{0} client-ip={1}; helo={2}; envelope-from={3}; receiver={4}'.format(comment, data.get('client_address', '<UNKNOWN>'),\
                                        data.get('helo_name', '<UNKNOWN>'), data.get('sender', '<UNKNOWN>'), data.get('recipient', '<UNKNOWN>'))
                                        )]))
                            except TypeError as x:
                                if str(x).startswith('sequence item 2'):
                                    syslog.syslog(syslog.LOG_ERR, 'Authserv_Id not set for authentication results header - invalid configuration.')
                                    raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                        else:
                            Header = ('X-Comment: ' + comment + 'client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                                % ( data.get('client_address', '<UNKNOWN>'),
                                    data.get('helo_name', '<UNKNOWN>'),
                                    data.get('sender', '<UNKNOWN>'),
                                    data.get('recipient', '<UNKNOWN>'),
                                    ))
                        if debugLevel >= 3: syslog.syslog(Header)
                        return (True, Header)
            return (False, 'None')
#############################################
def _rejectmessage(result, type, info, ip, recipient, configData):
    if result[3] == 'reject':
        rejectdefer = "rejected"
    elif result[3] == 'defer':
        rejectdefer = "deferred"
    url = ("http://www.openspf.net/Why?s={0};id={1};ip={2};r={3}"
              .format(type, info, ip, recipient))
    msg = configData.get('Reason_Message')
    return msg.format(
            rejectdefer=rejectdefer,
            spf=result[1],
            url=url,
    )

#############################################
def _spfcheck(data, instance_dict, configData, peruser, peruserconfigData):  #{{{1
    debugLevel = configData.get('debugLevel', 1)
    if not peruser:
        Void_Limit = configData.get('Void_Limit')
    else:
        Void_Limit = peruserconfigData.get('Void_Limit')
    spf.MAX_VOID_LOOKUPS = Void_Limit
    if configData.get('Prospective'):
        ip = configData.get('Prospective')
    else:
        ip = data.get('client_address')
    if ip == None:
        if debugLevel >= 2: syslog.syslog('spfcheck: No client address, exiting')
        return(( None, None, instance_dict, None))
    instance = data.get('instance')
    # The following if is only needed for testing.  Postfix 
    # will always provide instance.
    if not instance:
        import random
        instance = str(int(random.random()*100000))
    # This is to prevent multiple headers being prepended
    # for multi-recipient mail.
    if instance in instance_dict:
        found_instance = instance_dict[instance]
    else:
        found_instance = []
    # If this is not the first recipient for the message, we need to know if
    # there is a previous prepend to make sure we don't prepend more than once.
    if found_instance:
        if found_instance[5] != 'prepend':
            last_action = found_instance[3]
        else:
            last_action = found_instance[5]
    else:
        last_action = ""
    #  start query
    spfResult = None
    spfReason = None
    '''Data structure for results is a list of:
        [0] SPF result 
        [1] SPF reason
        [2] Identity (HELO/Mail From)
        [3] Action based on local policy
        [4] Header
        [5] last_action (need to know if we've prepended already)'''
    if debugLevel >= 4: syslog.syslog('Cached data for this instance: %s' % str(found_instance))
    if not found_instance or peruser:
        # Do not check SPF for localhost addresses
        if not peruser:
            skip_check = _spfbypass(data, 'skip_addresses', configData)
        else:
            skip_check = _spfbypass(data, 'skip_addresses', peruserconfigData)
            if debugLevel >= 4: syslog.syslog('skip_addresses check used peruser data')
        if skip_check[0]:
            skip_data = ["N/A", "Skip SPF checks on localhost", "N/A", "prepend", skip_check[1]]
            if last_action != 'prepend':
                skip_data.append('prepend')
                if not found_instance and not peruser:
                    instance_dict[instance] = skip_data
                return (('prepend', skip_check[1], instance_dict, None))
            else:
                return(( 'dunno', 'Header already pre-pended', instance_dict, None))
        # Whitelist designated IP addresses from SPF checks (e.g. secondary MX or 
        # known forwarders.
        if not peruser:
            ip_whitelist = _spfbypass(data, 'Whitelist', configData)
        else:
            ip_whitelist = _spfbypass(data, 'Whitelist', peruserconfigData)
            if debugLevel >= 4: syslog.syslog('Whitelist check used peruser data')
        if ip_whitelist:
            if ip_whitelist[0]:
                skip_data = ["N/A", "Skip SPF checks for whitelisted IP addresses", "N/A", "prepend", skip_check[1]]
                if last_action != 'prepend':
                    skip_data.append('prepend')
                    if not found_instance and not peruser:
                        instance_dict[instance] = skip_data
                if last_action != 'prepend':
                    return (('prepend', ip_whitelist[1], instance_dict, None))
                else:
                    return(( 'dunno', 'Header already pre-pended', instance_dict, None))
        # Whitelist designated Domain's sending addresses from SPF checks (e.g. 
        # known forwarders.
        if not peruser:
            domain_whitelist = _spfbypass(data, 'Domain_Whitelist', configData)
        else:
            domain_whitelist = _spfbypass(data, 'Domain_Whitelist', peruserconfigData)
            if debugLevel >= 4: syslog.syslog('Domain_Whitelist check used peruser data')
        if domain_whitelist:
            if domain_whitelist[0]:
                skip_data = ["N/A", "Skip SPF checks for whitelisted domains", "N/A", "prepend", skip_check[1]]
                if last_action != 'prepend':
                    skip_data.append('prepend')
                    if not found_instance and not peruser:
                        instance_dict[instance] = skip_data
                if last_action != 'prepend':
                    return (('prepend', domain_whitelist[1], instance_dict, None))
                else:
                    return(( 'dunno', 'Header already pre-pended', instance_dict, None))
        # Whitelist designated HELO/EHLO Domain's sending addresses from SPF
        # checks (e.g. known forwarders and hosts with broken SPF records).
        if not peruser:
            helo_whitelist = _spfbypass(data, 'HELO_Whitelist', configData)
        else:
            helo_whitelist = _spfbypass(data, 'HELO_Whitelist', peruserconfigData)
            if debugLevel >= 4: syslog.syslog('HELO_Whitelist check used peruser data')
        if helo_whitelist:
            if helo_whitelist[0]:
                skip_data = ["N/A", "Skip SPF checks for whitelisted HELO/EHLO names", "N/A", "prepend", skip_check[1]]
                if last_action != 'prepend':
                    skip_data.append('prepend')
                    if not found_instance and not peruser:
                        instance_dict[instance] = skip_data
                if last_action != 'prepend':
                    return (('prepend', helo_whitelist[1], instance_dict, None))
                else:
                    return(( 'dunno', 'Header already pre-pended', instance_dict, None))
        # Whitelist designated Domain's sending addresses from SPF checks (e.g.
        # known forwarders, but based on PTR match
        if not peruser:
            domain_whitelist = _spfbypass(data, 'Domain_Whitelist_PTR', configData)
        else:
            domain_whitelist = _spfbypass(data, 'Domain_Whitelist_PTR', peruserconfigData)
        if domain_whitelist:
            if domain_whitelist[0]:
                skip_data = ["N/A", "Skip SPF checks for whitelisted hosts (by PTR)", "N/A", "prepend", skip_check[1]]
                if last_action != 'prepend':
                    skip_data.append('prepend')
                    if not found_instance and not peruser:
                        instance_dict[instance] = skip_data
                if last_action != 'prepend':
                    return (('prepend', domain_whitelist[1], instance_dict, None))
                else:
                    return(( 'dunno', 'Header already pre-pended', instance_dict, None))
        receiver=socket.gethostname()
        sender = data.get('sender')
        helo = data.get('helo_name')
        if not sender and not helo:
            if debugLevel >= 2: syslog.syslog('spfcheck: No sender or helo, exiting')
            return(( None, None, instance_dict, None))
        if not peruser:
            Mail_From_pass_restriction = configData.get('Mail_From_pass_restriction')
            HELO_pass_restriction = configData.get('HELO_pass_restriction')
            HELO_reject = configData.get('HELO_reject')
            Mail_From_reject = configData.get('Mail_From_reject')
            if configData.get('No_Mail') == 'True':
                No_Mail = True
            else: No_Mail = False
        else:
            Mail_From_pass_restriction = peruserconfigData.get('Mail_From_pass_restriction')
            HELO_pass_restriction = peruserconfigData.get('HELO_pass_restriction')
            HELO_reject = peruserconfigData.get('HELO_reject')
            Mail_From_reject = peruserconfigData.get('Mail_From_reject')
            if peruserconfigData.get('No_Mail') == 'True':
                No_Mail = True
            else: No_Mail = False
        helo_result = ['None',]
        # First do HELO check
        #  if no helo name sent, use domain from sender for later use.
        if not helo:
            foo = sender.split('@', 1)
            if len(foo) <  2: helo = 'unknown'
            else: helo = foo[1]
        if HELO_reject != 'No_Check':
            helo_fake_sender = 'postmaster@' + helo
            heloquery = spf.query(i=ip, s=helo_fake_sender, h=helo, querytime=configData.get('Lookup_Time'))
            try:
                res = heloquery.check()
            except Exception as e:
                e = sys.exc_info()
                exceptionmessage = "Exception: %s, locals: %s" %(e, locals())
                syslog.syslog("Ouch, caught exc: %s" %exceptionmessage)
                return(( 'dunno', exceptionmessage, instance_dict, None))
            helo_result = [res[0], res[2]]
            helo_result.append('helo') 
            helo_result[0] = helo_result[0].lower()
            helo_result[0] = helo_result[0].capitalize()
            if (helo_result[0] == 'Permerror') or (helo_result[0] == 'Temperror'):
                isheloerror = True
            else:
                isheloerror = False
            if not peruser:
                helo_resultpolicy, local = _get_resultcodes(data, configData, 'helo')
            else:
                helo_resultpolicy, local = _get_resultcodes(data, peruserconfigData, 'helo')
            if debugLevel >= 2:
                syslog.syslog('spfcheck: pyspf result: "%s"' % str(helo_result))
            if HELO_reject == 'Null' and sender:
                helo_result.append('dunno')
            else:
                for poss_actions in helo_resultpolicy:
                    if helo_result[0] in helo_resultpolicy[poss_actions]:
                        action = poss_actions
                        helo_result.append(action)
                if local['local_helo']:
                    helo_result[1] = 'Receiver policy for SPF ' + helo_result[0]
            if sender == '':
                header_sender = '<>'
            else:
                header_sender = sender
            if helo_result[0] == 'None':
                helo_result[1] = "no SPF record"
            spfDetail = ('identity=%s; client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                % (helo_result[2], ip, helo, header_sender, data.get('recipient', '<UNKNOWN>')))
            if debugLevel >= 2:
                logdata = str(helo_result[0]) + "; " + spfDetail
                syslog.syslog(logdata)
            header = ''
            if 'SPF' in configData.get('Header_Type') and 'AR' in configData.get('Header_Type'):
                if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR, 'Header_Type includes both SPF and AR - invalid configuration.')
                raise SyntaxError('Header_Type includes both SPF and AR - invalid configuration.')
            if 'AR' in configData.get('Header_Type'):
                if not configData.get('Authserv_Id'):
                    raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                header += str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                    results = [authres.SPFAuthenticationResult(result = helo_result[0],
                    result_comment = helo_result[1],
                    smtp_helo = helo, smtp_helo_comment =
                    'client-ip={0}; helo={1}; envelope-from={2}; receiver={3}'.format(ip, helo, header_sender, data.get('recipient', '<UNKNOWN>')))]))
            else:
                header = 'Received-SPF: '+ helo_result[0] + ' (' + helo_result[1] +') ' + spfDetail
            if helo_result[3] != 'reject' and helo_result[3] != 'defer':
                helo_result.append(header)
                helo_result.append(helo_result[3])
                if HELO_pass_restriction and helo_result[0] == 'Pass':
                    restrict_name = HELO_pass_restriction
                    helo_result[3] = 'result_only'
                    helo_result[4] = restrict_name
                    helo_result[5] = 'restriction'
                if not peruser:
                    instance_dict[instance] = helo_result
            # Only act on the HELO result if it is authoritative.
            else:
                if helo_result[3] == 'reject':
                    if No_Mail:
                        # If only rejecting on "v=spf1 -all", we need to know now
                        if debugLevel >= 5: syslog.syslog('Pyspf Cached SPF record before HELO No_Mail check: {0}'.format(str(heloquery.cache[helo, 'TXT'][0][0])))
                        record = str(heloquery.cache[helo, 'TXT'][0][0])
                        if record != str("b'v=spf1 -all'"):
                            if last_action != 'prepend':
                                # Prepend instead of reject if it's not a no mail record
                                helo_result[3] = 'prepend'
                            else:
                                return(( 'dunno', 'Header already pre-pended', instance_dict, None))
                if helo_result[3] in ('reject', 'defer'): # It may not be reject anymore
                    header = _rejectmessage(helo_result, 'helo', helo, ip, data.get('recipient'), configData)
                    helo_result.append(header)
                    helo_result.append(helo_result[3])
                    if not peruser:
                        instance_dict[instance] = helo_result
                    return((helo_result[3], header, instance_dict, isheloerror ))
        # Second do Mail From Check
        if sender == '':
            if HELO_reject != 'No_Check':
                helo_result.append(header)
                helo_result.append(helo_result[3])
                if HELO_pass_restriction and helo_result[0] == 'Pass':
                    restrict_name = HELO_pass_restriction
                    helo_result[3] = 'result_only'
                    helo_result[4] = restrict_name
                    helo_result[5] = 'restriction'
                    if not peruser:
                        instance_dict[instance] = helo_result
                    return('result_only', restrict_name, instance_dict, None)
                if not peruser:
                    instance_dict[instance] = helo_result
                return(( helo_result[3], header, instance_dict, isheloerror ))
        else:
            if Mail_From_reject != 'No_Check':
                mfromquery = spf.query(i=ip, s=sender, h=helo, querytime=configData.get('Lookup_Time'))
                try:
                    mres = mfromquery.check()
                except Exception as e:
                    e = sys.exc_info()
                    exceptionmessage = "Exception: %s, locals: %s" %(e, locals())
                    syslog.syslog("Ouch, caught exc: %s" %exceptionmessage)
                    return(( 'dunno', exceptionmessage, instance_dict, None))
                mfrom_result = [mres[0], mres[2]]
                mfrom_result.append('mailfrom')
                mfrom_result[0] = mfrom_result[0].lower()
                mfrom_result[0] = mfrom_result[0].capitalize()
                if (mfrom_result[0] == 'Permerror') or (mfrom_result[0] == 'Temperror'):
                    ismfromerror = True
                else:
                    ismfromerror = False
                if not peruser:
                    mfrom_resultpolicy, local = _get_resultcodes(data, configData, 'mfrom')
                else:
                    mfrom_resultpolicy, local = _get_resultcodes(data, peruserconfigData, 'mfrom')
                if debugLevel >= 2:
                    syslog.syslog('spfcheck: pyspf result: "%s"' % str(mfrom_result))
                for poss_actions in mfrom_resultpolicy:
                    if mfrom_result[0] in mfrom_resultpolicy[poss_actions]:
                        action = poss_actions
                        mfrom_result.append(action)
                if local['local_mfrom']:
                    mfrom_result[1] = 'Receiver policy for SPF ' + mfrom_result[0]
                if mfrom_result[0] == 'None':
                    mfrom_result[1] = 'no SPF record'
                if mfrom_result[0] != 'None' or (mfrom_result[0] == 'None' and helo_result[0] == 'None'):
                    spfDetail = \
                        ('identity=%s; client-ip=%s; helo=%s; envelope-from=%s; receiver=%s '
                        % (mfrom_result[2], ip, helo, sender, data.get('recipient', '<UNKNOWN>')))
                    if debugLevel >= 2:
                        logdata = str(mfrom_result[0]) + "; " + spfDetail
                        syslog.syslog(logdata)
                    header = ''
                    if 'SPF' in configData.get('Header_Type') and 'AR' in configData.get('Header_Type'):
                        if debugLevel >= 0: syslog.syslog(syslog.LOG_ERR, 'Header_Type includes both SPF and AR - invalid configuration.')
                        raise SyntaxError('Header_Type includes both SPF and AR - invalid configuration.')
                    if 'AR' in configData.get('Header_Type'):
                        if not configData.get('Authserv_Id'):
                            raise SyntaxError('Authserv_Id not set for authentication results header - invalid configuration.')
                        header += str(authres.AuthenticationResultsHeader(authserv_id = configData.get('Authserv_Id'),
                            results = [authres.SPFAuthenticationResult(result = mfrom_result[0],
                            result_comment = mfrom_result[1],
                            smtp_mailfrom = spf.split_email(sender,'example.com')[1], smtp_mailfrom_comment =
                            'client-ip={0}; helo={1}; envelope-from={2}; receiver={3}'.format(ip, helo, sender, data.get('recipient', '<UNKNOWN>')))]))
                    else:
                        header = 'Received-SPF: '+ mfrom_result[0] + ' (' + mfrom_result[2] +') ' + spfDetail
                    if mfrom_result[3] != 'reject' and mfrom_result[3] != 'defer':
                        mfrom_result.append(header)
                        mfrom_result.append(mfrom_result[3])
                        if not peruser:
                            if debugLevel >= 4:syslog.syslog('not peruser')
                            instance_dict[instance] = mfrom_result
                if (Mail_From_pass_restriction and mfrom_result[0] == 'Pass') or \
                   (HELO_pass_restriction and helo_result[0] == 'Pass'):
                    if mfrom_result[0] == 'Pass':
                        restrict_name = Mail_From_pass_restriction
                        mfrom_result[3] = 'result_only'
                        mfrom_result[4] = restrict_name
                        mfrom_result[5] = 'restriction'
                        if not peruser:
                            instance_dict[instance] = mfrom_result
                    return('result_only', restrict_name, instance_dict, None)
                # Act on the Mail From result if it is authoritative.
                if mfrom_result[3] == 'reject' or mfrom_result[3] == 'defer':
                    if mfrom_result[3] == 'reject':
                        if No_Mail:
                            # If only rejecting on "v=spf1 -all", we need to know now
                            senderdomain = sender.split('@')[1]
                            if debugLevel >= 5: syslog.syslog('Pyspf Cached SPF record before mfrom No_Mail check: {0}'.format(str(mfromquery.cache[senderdomain, 'TXT'][0][0])))
                            mrecord = str(mfromquery.cache[senderdomain, 'TXT'][0][0])
                            if mrecord != str("b'v=spf1 -all'"):
                                if last_action != 'prepend':
                                    # Prepend instead of reject if it's not a no mail record
                                    mfrom_result[3] = 'prepend'
                                else:
                                    return(( 'dunno', 'Header already pre-pended', instance_dict, None ))
                    if mfrom_result[3] in ('reject', 'defer'): # It may not be reject anymore
                        header = _rejectmessage(mfrom_result, 'mfrom', sender, ip, data.get('recipient'), configData)
                        mfrom_result.append(header)
                        mfrom_result.append(mfrom_result[3])
                        if not peruser:
                            instance_dict[instance] = mfrom_result
                        return(( mfrom_result[3], header, instance_dict, ismfromerror ))
                if mfrom_result[3] != 'dunno' or helo_result[3] == 'dunno':
                    if last_action != 'prepend':
                        return(( 'prepend', header, instance_dict, None ))
                    else:
                        return(( 'dunno', 'Header already pre-pended', instance_dict, None ))
            else:
                if HELO_reject != 'No_Check':
                    if helo_result[3] == 'reject':
                        if No_Mail:
                            # If only rejecting on "v=spf1 -all", we need to know now
                            if debugLevel >= 5: syslog.syslog('Pyspf Cached SPF record before HELO No_Mail check: {0}'.format(str(heloquery.cache[helo, 'TXT'][0][0])))
                            record = str(heloquery.cache[helo, 'TXT'][0][0])
                            if record != str("b'v=spf1 -all'"):
                                if last_action != 'prepend':
                                    # Prepend instead of reject if it's not a no mail record
                                    helo_result[3] = 'prepend'
                                else:
                                    return(( 'dunno', 'Header already pre-pended', instance_dict, None ))
                    if helo_result[3] in ('reject', 'defer'): # It may not anymore
                        header = _rejectmessage(helo_result, 'helo', helo, ip, data.get('recipient'), configData)
                        if helo_result[3] == 'reject': # It may not anymore
                            return(( 'reject', header, instance_dict, isheloerror ))
                        if helo_result[3] == 'defer':
                            return(( 'defer', header, instance_dict, isheloerror ))
                    if last_action != 'prepend':
                        return(( 'prepend', header, instance_dict, None ))
                    else:
                        return(( 'dunno', 'Header already pre-pended', instance_dict, None ))
    else:
        cached_instance = instance_dict[instance]
        if cached_instance[3] == 'prepend':
            return(( 'dunno', 'Header already pre-pended', instance_dict, None ))
        else:
            return(( cached_instance[3], cached_instance[4], instance_dict, None ))
    return(( 'None', 'None', instance_dict, None ))




from setuptools import setup
import os
import sys

DESC = """SPF (Sender Policy Framework) processing engine for Postfix policy server and Milter implemented in Python."""

setup(name='spf-engine',
    version='2.9.3',
    description='SPF processing for Postfix (and Sendmail)',
    long_description=DESC,
    author='Scott Kitterman',
    author_email='scott@kitterman.com',
    url='https://launchpad.net/spf-engine',
    packages=['spf_engine'],
    keywords = ['Postfix', 'Sendmail', 'milter','spf','email'],
    zip_safe = False,
    install_requires = ['pyspf',],
    extras_require={'milter':  ['pymilter', 'authres',],
    },
    entry_points = {
        'console_scripts' : [
            'policyd-spf = spf_engine.policyd_spf:main',
            'pyspf-milter = spf_engine.milter_spf:main',
        ]},
    include_package_data=True,
    data_files=[(os.path.join('share', 'man', 'man1'),
        ['policyd-spf.1']), (os.path.join('share', 'man', 'man5'),
        ['policyd-spf.conf.5']), (os.path.join('etc', 'python-policyd-spf'),
        ['policyd-spf.conf']), (os.path.join('share', 'man', 'man5'),
        ['policyd-spf.peruser.5']), (os.path.join('lib', 'systemd', 'system'),
        ['system/pyspf-milter.service']),(os.path.join('etc', 'init.d'),
        ['system/pyspf-milter']) ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email :: Mail Transport Agents',
        'Topic :: Communications :: Email :: Filters',
    ]
)

if sys.version_info < (3, 3):
    raise Exception("SPF engine requires python3.3 and later.")



# drop_priviledges (from https://github.com/nigelb/Static-UPnP)
# Copyright (C) 2016  NigelB
# Copyright (C) 2018 Scott Kitterman
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


def fold(header):
    """Fold a header line into multiple crlf-separated lines at column 72.
    Borrowed from dkimpy and updated to only add \n instead of \r\n because
    that's what the milter protocol wants.

    >>> text(fold('foo'))
    'foo'
    >>> text(fold('foo  '+'foo'*24).splitlines()[0])
    'foo  '
    >>> text(fold('foo'*25).splitlines()[-1])
    ' foo'
    >>> len(fold('foo'*25).splitlines()[0])
    72
    """
    i = header.rfind("\r\n ")
    if i == -1:
        pre = ""
    else:
        i += 3
        pre = header[:i]
        header = header[i:]
    maxleng = 72
    while len(header) > maxleng:
        i = header[:maxleng].rfind(" ")
        if i == -1:
            j = maxleng
        else:
            j = i + 1
        pre += header[:j] + "\n   "
        header = header[j:]
    return pre + header


def user_group(userid):
    """Return user and group from UserID"""
    import grp
    import pwd

    userlist = userid.split(':')
    if len(userlist) == 1:
        gidname = userlist[0]
    else:
        gidname = userlist[1]
    # Get the uid/gid from the name
    running_uid = pwd.getpwnam(userlist[0]).pw_uid
    running_gid = grp.getgrnam(gidname).gr_gid
    return running_uid, running_gid


def drop_privileges(milterconfig):
    import os
    import syslog

    if os.getuid() != 0:
        if milterconfig.get('Syslog'):
            syslog.syslog('drop_privileges: Not root. No action taken.')
        return

    # Get user and group
    uid, gid = user_group(milterconfig.get('UserID'))

    # Remove group privileges
    os.setgroups([])

    # Try setting the new uid/gid
    os.setgid(gid)
    os.setuid(uid)

    # Set umask
    old_umask = os.umask(milterconfig.get('UMask'))


def write_pid(milterconfig):
    """Write PID in pidfile.  Will not overwrite an existing file."""
    import os
    import syslog
    if not os.path.isfile(milterconfig.get('PidFile')):
        pid = str(os.getpid())
        try:
            f = open(milterconfig.get('PidFile'), 'w')
        except IOError as e:
            if str(e)[:35] == '[Errno 2] No such file or directory':
                piddir = milterconfig.get('PidFile').rsplit('/', 1)[0]
                os.mkdir(piddir)
                user, group = user_group(milterconfig.get('UserID'))
                os.chown(piddir, user, group)
                f = open(milterconfig.get('PidFile'), 'w')
                if milterconfig.get('Syslog'):
                    syslog.syslog('PID dir created: {0}'.format(piddir))
            else:
                if milterconfig.get('Syslog'):
                    syslog.syslog('Unable to write pidfle {0}.  IOError: {1}'
                                  .format(milterconfig.get('PidFile'), e))
                raise
        f.write(pid)
        f.close()
        user, group = user_group(milterconfig.get('UserID'))
        os.chown(milterconfig.get('PidFile'), user, group)
    else:
        if milterconfig.get('Syslog'):
            syslog.syslog('Unable to write pidfle {0}.  File exists.'
                          .format(milterconfig.get('PidFile')))
        raise RuntimeError('Unable to write pidfle {0}.  File exists.'
                           .format(milterconfig.get('PidFile')))
    return pid


